
open Types

  
let output_string oc s =
(*  Printf.printf "Sending to server:%s" s;  print_newline ();  *)
  output_string oc s
  
module Make(Strategy: sig
      val init : init -> unit
        
      val play : state -> int * command
        
      val busy_before_packages : (state -> unit) option
      val busy_before_reply : (state -> unit) option
    end) = struct
  
(*********************************************************)
(*                                                       *)
(*      Functions to communicate with server             *)
(*                                                       *)
(*********************************************************)

let read_init ic =

(* First of all, read the map *)
  let coords = input_line ic in
  let (dx,dy) = match String2.split_simplify coords ' ' with
      [dx;dy] -> int_of_string dx, int_of_string dy
    | _ -> 
        Printf.printf "Buggy first line from server: %s" coords; 
        print_newline ();
        assert false
  in
(*  Printf.printf "READING MAP %d %d" dx dy; print_newline (); *)
  let map = Array.make_matrix (dx+2) (dy+2) Wall in
  for y = 1 to dy do
    let line = input_line ic in
    for x = 1 to dx do
      map.(x).(y) <- 
      match line.[x-1] with
        '.' -> EmptyUnchecked
      | '@' -> HomeUnchecked
      | '~' -> SeaUnchecked
      | '#' -> Wall
      | c -> 
          Printf.printf "Bad char '%c' in line %s" c line; print_newline ();
          assert false
    done
  done;
(*  Printf.printf "MAP READ"; print_newline (); *)
  let my_line = input_line ic in
  let my_num, weight, money = match String2.split_simplify my_line ' ' with
      [num; weight; money] ->
        int_of_string num, int_of_string weight, int_of_string money
    | _ ->
        Printf.printf "Buggy first line in %s" my_line; print_newline ();
        assert false
  in
(*  Printf.printf "OWN INFO READ"; print_newline (); *)
  let status_line = input_line ic in
  let status_list = String2.split_simplify status_line ' ' in
  let nplayers = ref 0 in
  let rec iter1 list left =
    match list with
      [] -> assert false
    | num :: "X" :: x :: "Y" :: y :: tail ->
        let len = String.length num in
        let num = String.sub num 1 (len-1) in
(*        Printf.printf "[%s] [%s] [%s]" num x y; print_newline (); *)
        let num = int_of_string num in
        let x = int_of_string x in
        let y = int_of_string y in
        if num = my_num then iter2 x y tail left
        else
        iter1 tail ((num, x, y) :: left)
    | _ -> 
        Printf.printf "Unparsable status line from server"; print_newline ();
        assert false
  and  iter2 xx yy list left =
    match list with
      [] -> xx, yy, left
    | num :: "X" :: x :: "Y" :: y :: tail ->
        let len = String.length num in
        let num = String.sub num 1 (len-1) in
(*        Printf.printf "[%s] [%s] [%s]" num x y; print_newline (); *)
        iter2 xx yy tail 
          ((int_of_string num, int_of_string x, int_of_string y) :: left)
    | _ -> 
        Printf.printf "Unparsable status line from server"; print_newline ();
        assert false
  in
  let x , y, players = iter1 status_list [] in
  {
    init_map = map;
    init_dx = dx;
    init_dy = dy;
    init_num = my_num;
    init_money = money;
    init_x = x;
    init_y = y;
    init_weight = weight;
    init_nplayers = !nplayers;
    init_players = players;
    init_nsea = 0;
    init_nhome = 0;
    init_nempty = 0;
    init_homes = [];
  }

let string_of_move (bid, move) =
  let buf = Buffer.create 1000 in
  Printf.bprintf buf "%d " bid;
  begin
    match move with
      Pick list -> 
        Buffer.add_string buf "Pick ";
        List.iter (fun p -> Printf.bprintf buf "%d " p.pack_num) list;
    | Drop list -> 
        Buffer.add_string buf "Drop ";
        List.iter (fun p -> Printf.bprintf buf "%d " p.pack_num ) list;
    | Move dir ->
        Buffer.add_string buf "Move ";
        Buffer.add_char buf (match dir with
            E -> 'E'  
          | W -> 'W'
          | N -> 'N'  
          | S -> 'S')
    | XY _ -> assert false
  end;
  Buffer.add_char buf '\n';
  Buffer.contents buf

let rec iter_moves list moves =
  match list with 
    [] -> List.rev moves
  | num :: tail ->
      assert (num.[0] = '#');
      let len = String.length num in
      let num = int_of_string (String.sub num 1 (len-1)) in
      iter_player_moves tail moves num []
      
and iter_player_moves list moves num pmoves =
  match list with
    [] ->
      List.rev  ((num, List.rev pmoves) :: moves)
  | "P" :: pid :: tail ->
      iter_player_moves tail moves num ((Pick [new_package (int_of_string pid)]) :: pmoves)
  | "D" :: pid :: tail ->
      iter_player_moves tail moves num ((Drop [new_package (int_of_string pid)]) :: pmoves)
  | "E" :: tail ->
      iter_player_moves tail moves num ((Move E) :: pmoves)
  | "N" :: tail ->
      iter_player_moves tail moves num ((Move N) :: pmoves)
  | "S" :: tail ->
      iter_player_moves tail moves num ((Move S) :: pmoves)
  | "W" :: tail ->
      iter_player_moves tail moves num ((Move W) :: pmoves)
  | "X" :: x :: "Y" :: y :: tail ->
      iter_player_moves tail moves num
        ((XY (int_of_string x, int_of_string y)) :: pmoves)
  | _ ->
      iter_moves list ((num, List.rev pmoves) :: moves)

  
let connect_server host port =
  let sock = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
  let h = Unix.gethostbyname host in
  let ip = h.Unix.h_addr_list.(0) in
  let sockaddr = Unix.ADDR_INET (ip, port) in
  Unix.connect sock sockaddr;
  Printf.printf "Connected to %s:%d" host port; print_newline ();
  let oc = Unix.out_channel_of_descr sock in
  let ic = Unix.in_channel_of_descr sock in
  sock, ic, oc  

let busyp fd f state =
  match f with
    None -> ()
  | Some f ->
      while
        match Unix.select [fd] [] [] 0.0 with
          [], [], [] -> true
        | _ -> false
      do
        f state
      done

(*********************************************************)
(*                                                       *)
(*      Simplification of the Map                        *)
(*                                                       *)
(*********************************************************)

let simplify_map init = 
  let nhome = ref 0 in
  let nsea = ref 0 in
  let nempty = ref 0 in
  let map = init.init_map in
  let homes = ref [] in
  let rec iter x y =
    match map.(x).(y) with
    | HomeUnchecked ->
        map.(x).(y) <- Home;
        homes := (x,y) :: !homes;
        incr nhome;
        iter x (y+1);
        iter x (y-1);
        iter (x+1) y;
        iter (x-1) y
    | EmptyUnchecked ->
        map.(x).(y) <- Empty;
        incr nempty;
        iter x (y+1);
        iter x (y-1);
        iter (x+1) y;
        iter (x-1) y        
    | SeaUnchecked ->
        map.(x).(y) <- Sea;
        incr nsea
    | _ -> ()
  in
  iter init.init_x init.init_y;
  for x = 1 to init.init_dx do
    for y = 1 to init.init_dy do
      match map.(x).(y) with
        HomeUnchecked | EmptyUnchecked | SeaUnchecked ->
          map.(x).(y) <- Wall
      | _ -> ()
    done
  done;
  init.init_nsea <- !nsea;
  init.init_nhome <- !nhome;
  init.init_homes <- !homes;
  init.init_nempty <- !nempty;
  ()
  
  
(*********************************************************)
(*                                                       *)
(*                  Arguments                            *)
(*                                                       *)
(*********************************************************)

let host = ref "localhost"
let port = ref 20005
  

      let _ = begin match Sys.argv with [| _; h; p |] -> host := h;
        port := int_of_string p | _ -> () end;

  
  
(*********************************************************)
(*                                                       *)
(*                  Main                                 *)
(*                                                       *)
(*********************************************************)
  
  let fd, ic,oc = connect_server !host !port in
  output_string oc "Player\n";
  flush oc;
  
  let init = read_init ic in

  simplify_map init;
  Strategy.init init;

  let rec iter_packages line list left =
    match list with
      [] -> left
    | num :: final_x :: final_y :: weight :: tail ->
        
        iter_packages line tail 
          (let p = new_package (int_of_string num) in
          p.pack_final <-  Some (int_of_string final_x, int_of_string final_y);
          p.pack_weight <- int_of_string weight;
          p  :: left)
    | _ -> 
        Printf.printf "Bad package line from server %s" line; 
        print_newline ();        
        assert false
  in

  let state = {
      state_map = init.init_map;
      state_location_packages = [];
      state_moves = [];
    } in
  
  while true do

    busyp fd Strategy.busy_before_packages state;    

    let line = input_line ic in
    
    let packages = iter_packages line (String2.split_simplify line ' ') [] in
    
    state.state_location_packages <- packages;
    
    let move = Strategy.play state in
    output_string oc (string_of_move move);
    flush oc;

    state.state_moves <- [];
    
    busyp fd Strategy.busy_before_reply state;
    
    let line = input_line ic in
    let moves = String2.split_simplify line ' ' in

    let moves = iter_moves moves [] in
    state.state_moves <- moves;
        
  done
  
end
