open Bigarray
open Types

type current_state =
  Looking_for_packages
| MoveToHome of home
| MoveToPackage of package
  
module CoordSet =
  Set.Make(struct type t = int * int let compare = compare end)
  
let dists = Hashtbl.create 997
let dists_dom = ref CoordSet.empty
    
let dists_aux = ref (Array2.create int c_layout 0 0)
let map = ref [|[||]|]

let num = ref 0

let x = ref 0
let y = ref 0
let dx = ref 0
let dy = ref 0
let homes = ref []
let packages = ref  [|[||]|]
let current_state = ref Looking_for_packages

let total_weight = ref 0
let current_weight = ref 0
let current_packages = ref []

let score = ref 0
  
let compute_distances_from (x0,y0 as pt0) =
  let worklist = Queue.create () in
  let aux = !dists_aux in
  Array2.fill aux max_int;
  aux.{x0,y0} <- 0;
  Queue.add (x0,y0) worklist;
  let update x y d =
    match !map.(x).(y) with
      Empty | Home ->
        if d < aux.{x,y} then begin
          aux.{x,y} <- d;
          Queue.add (x,y) worklist
        end
    | _ -> () in
  begin try
    while true do
      let (x,y) = Queue.take worklist in
      let d = aux.{x,y} + 1 in
      update (x+1) y d;
      update (x-1) y d;
      update x (y+1) d;
      update x (y-1) d
    done
  with Queue.Empty ->
    ()
  end;
  CoordSet.iter
    (fun (x,y as pt) ->
      let d = aux.{x,y} in
      let c = compare pt0 pt in
      if c < 0 then Hashtbl.add dists (x0,y0,x,y) d
      else if c > 0 then Hashtbl.add dists (x,y,x0,y0) d)
    !dists_dom

(* Return distance between two points.  Cache the results. *)

let rec dist (x1,y1 as pt1) (x2,y2 as pt2) =
  try
    let c = compare pt1 pt2 in
    if c = 0 then 0
    else if c < 0 then Hashtbl.find dists (x1,y1,x2,y2)
    else Hashtbl.find dists (x2,y2,x1,y1)
  with Not_found ->
    dists_dom := CoordSet.add pt1 (CoordSet.add pt2 !dists_dom);
    compute_distances_from pt1;
    dist pt1 pt2

let sort_path () =
  match !current_packages with
    [] -> ()
  | ({ pack_final = Some(xx,yy) } as p) :: tail ->
      let shortest_dist = ref (dist (xx,yy) (!x,!y))  in
      let best_dest = ref p in
      List.iter (fun p ->
          match p.pack_final with
            Some(xx,yy) ->
              let d = dist (xx,yy) (!x,!y) in
              if !shortest_dist > d then begin
                  best_dest := p;
                  shortest_dist := d;
                end
          | _ -> ()
      ) tail;
      current_packages := !best_dest :: (
        List2.removeq !best_dest !current_packages)
  | _ -> assert false

let close_sea x y =
  !map.(x+1).(y) = Sea || 
  !map.(x-1).(y) = Sea || 
  !map.(x).(y+1) = Sea || 
  !map.(x).(y-1) = Sea ||
  !map.(x+1).(y-1) = Sea ||
  !map.(x+1).(y+1) = Sea ||
  !map.(x-1).(y-1) = Sea ||
  !map.(x-1).(y+1) = Sea 
      
let move_dist x y xx yy =
    (match !map.(x).(y) with
      Home | Empty -> 
        (if close_sea x y then 1 else 0) + 2 * (dist (x,y) (xx,yy))
    |  _ -> 10000000)

let best_move_to xx yy =
  let north = move_dist !x (!y+1) xx yy in
  let south = move_dist !x (!y-1) xx yy in
  let east = move_dist (!x+1) !y xx yy in
  let west = move_dist (!x-1) !y xx yy in
  let best_dir = ref N in
  let best_dist = ref north in
  if south < north then (best_dir := S; best_dist := south);
  if east < !best_dist then (best_dir := E; best_dist := east);
  if west < !best_dist then (best_dir := W; best_dist := west);
  Move !best_dir

let update_state state =
  Hashtbl.iter (fun _ r ->  r.robot_alive <- false) robots;
      
  List.iter (fun (robot_num, list) ->
      if robot_num = !num then
        List.iter (fun command ->
            match command with
            | Move dir -> begin
                  match dir with
                    E -> incr x
                  | W -> decr x
                  | N -> incr y
                  | S -> decr y
                end;
(*                Printf.printf "New coord %d %d" !x !y; print_newline () *)
            | Pick list ->
                List.iter (fun p ->
                    current_weight := !current_weight + p.pack_weight;
(*                    Printf.printf "Picking %d for %d (final: %d/%d)"
                      p.pack_num p.pack_weight !current_weight !total_weight;
                    print_newline (); *)
                    current_packages := p :: !current_packages;
                    sort_path ();
                    p.pack_pos <- None;
                    p.pack_delivered <- false;
                ) list
                    
            | Drop list ->
                List.iter (fun p ->
                    p.pack_delivered <- true;
                    if p.pack_final = Some (!x,!y) then begin
                        Hashtbl.remove all_packages p.pack_num;
                        score := !score + p.pack_weight;
                        p.pack_pos <- None;
                      end else
                      p.pack_pos <- Some (!x,!y);
                    current_weight := !current_weight - p.pack_weight;
                    current_packages := List2.removeq p !current_packages;
                    sort_path ();
                ) list
                
            | _ -> ()
        ) list
      else
      let robot = new_robot robot_num in
      List.iter (fun command ->
          match command with
          | Move dir -> begin
                let (xx,yy) = robot.robot_pos in
                robot.robot_pos <- (match dir with
                    E -> (xx+1, yy)
                  | W -> (xx-1, yy)
                  | N -> (xx, yy+1)
                  | S -> (xx, yy-1))
              end;
          | Pick list -> 
              List.iter (fun p -> 
                  p.pack_delivered <- false;
                  p.pack_pos <- None) list;
              robot.robot_packages <- list @ robot.robot_packages
          | Drop list ->
              List.iter (fun p ->
                  let (rx,ry) = robot.robot_pos in
                  p.pack_pos <- Some (rx,ry);
                  if p.pack_pos = p.pack_final then begin
                      Hashtbl.remove all_packages p.pack_num;
                      p.pack_pos <- None;
                    end;
                  p.pack_delivered <- true;
                  robot.robot_packages <- List2.remove p robot.robot_packages
              ) list
          | XY (x,y) ->
              robot.robot_pos <- (x,y)
      ) list;
(*      Printf.printf "Robot %d: %d x %d" 
        robot.robot_num robot.robot_x robot.robot_y; 
      print_newline (); *)
    ) state.state_moves;
  if state.state_location_packages <> [] then begin
(*      Printf.printf "Availble packages:";  *)
      List.iter (fun p ->
(*          Printf.printf "%d " p.pack_num; 
          (match p.pack_final with
              None -> ()
            | Some (x,y) ->
                Printf.printf "(%d,%d) " x y;
          ); *)
          p.pack_pos <- Some (!x, !y);
      ) state.state_location_packages;
(*      print_newline (); *)
    end

let can_drop () = 
  
  let rec iter list to_drop to_keep =
    match list with
      p :: tail ->
        if p.pack_final = Some (!x, !y) then
          iter tail (p :: to_drop) to_keep
        else
          iter tail to_drop (p :: to_keep)
    | _ -> to_drop, to_keep
  in
  let (to_drop, to_keep) = iter !current_packages [] [] in
  to_drop

module CoordMap = Map.Make(struct type t = int * int let compare = compare end)

(* Find the closest location to a given path. *)
  
let choose_packets final_positions min_weight expected_move =
  let final_list = ref [] in
  CoordMap.iter (fun p f -> 
      final_list := f :: !final_list    
  ) final_positions;
  let expected_weight = ref !current_weight in
  
  let rec iter to_pick expected_move final_list =
(*    Printf.printf "iter"; print_newline (); *)
    let list = ref [] in
    let expected_dist = dist (!x,!y) expected_move in
    List.iter (fun f -> 
        let p = f.final_pos in
        let go_first = 
          (dist p (!x,!y) + dist p expected_move) - expected_dist in
        let go_after = dist p expected_move in
        
        f.final_after <-  go_first >= go_after;
        f.final_dist <- min go_first go_after;
        list := f :: !list) final_list;

(*    Printf.printf "sort"; print_newline (); *)
    list := List.sort (fun f1 f2 -> f1.final_dist - f2.final_dist) !list;
    
    let rec iter_in finals to_pick =
(*      Printf.printf "iter_in"; print_newline (); *)
      match finals with
        [] -> to_pick
      | f :: final_list ->
          let abort = ref false in
          let to_pick = ref to_pick in
          begin
            try
            List.iter (fun p ->
                if !expected_weight + min_weight > !total_weight then 
                  raise Exit;
(*                Printf.printf "Testing %d dist %d" p.pack_num f.final_dist; 
                print_newline (); *)
                if p.pack_weight <= !total_weight - !expected_weight then begin
                    to_pick := p :: !to_pick;
                    f.final_used <- true;
                    expected_weight := !expected_weight + p.pack_weight;
(*                    Printf.printf "Pick %d for %d obtaining %d/%d" p.pack_num
                      p.pack_weight !expected_weight !total_weight;
                    print_newline (); *)
                  end)
              f.final_packages;
            with Exit -> abort := true
          end;
          if !abort then !to_pick else
          if f.final_used && f.final_after then
            iter !to_pick f.final_pos final_list
          else
            iter_in  final_list !to_pick 
    in
    iter_in final_list to_pick
  in
  let v=iter [] expected_move !final_list in
(*  Printf.printf "iter done"; print_newline (); *)
  v
  
  
let can_pick state expected_move = 
  if !current_weight = !total_weight then [] else
    begin
      let final_positions = ref CoordMap.empty in
      let min_weight = ref max_int in
      List.iter (fun p ->
(*          Printf.printf "Possible %d of %d" p.pack_num p.pack_weight;
          print_newline (); *)
          match p.pack_final with
            None -> assert false
          | Some (xx,yy) ->
              if p.pack_weight <= !total_weight - !current_weight then
              match !map.(xx).(yy) with
              | Home | Empty ->
                  if p.pack_weight < !min_weight then min_weight := p.pack_weight;
                  begin
                    let f = try CoordMap.find (xx,yy) !final_positions
                      with _ ->
                            let f = {
                                final_after = false;
                                final_used = false;
                                final_pos = xx,yy;
                              final_packages =  [];
                                final_dist = dist (xx,yy) (!x,!y);
                                final_weight = 0;
                            } in
                          final_positions := CoordMap.add  (xx,yy) f  !final_positions;
                          f
                    in
                      f.final_packages <- p :: f.final_packages;
                      f.final_weight <- f.final_weight + p.pack_weight;
                  end
              | _ -> 
(*                  Printf.printf "Unaccessible"; print_newline ()  *)
                    ()
      ) state.state_location_packages;
            
      let list = ref [] in
      let to_pick = choose_packets !final_positions !min_weight expected_move in
      to_pick

    end
    
let is_dangerous dx dy =
  match !map.(!x+dx).(!y+dy) with
  | Empty | Home ->
      begin
        match !map.(!x+dx).(!y+dy) with
          Sea -> true
        | _ -> false
      end
  | _ -> false
    
let can_kill () =
  let x = !x in
  let y = !y in
  let dead_positions = 
    if is_dangerous 1 0 then [(x+1,y)] else [] in
  let dead_positions = 
    if is_dangerous (-1) 0 then 
      (x-11,y) :: dead_positions else dead_positions in
  let dead_positions = 
    if is_dangerous 0 1 then 
      (x, y+1) :: dead_positions else dead_positions in
  let dead_positions = 
    if is_dangerous 0 (-1) then
      (x, y-1) :: dead_positions else dead_positions in
  if dead_positions = [] then [] else
    let list = ref [] in
  Hashtbl.iter (fun _ r ->
      if r.robot_alive then
        if List.mem r.robot_pos dead_positions then
          list := r :: !list
  ) robots;
  !list
    
module M = Client.Make(struct

(* This function is called with the initial state received from the server *)
      let init ( init : init) = 
        map := init.init_map;
        dists_dom := CoordSet.empty;
        dists_aux := Array2.create int c_layout (init.init_dx+2) (init.init_dy+2);
        homes := List.map (fun (x,y) ->
(*            Printf.printf "home %d %d" x y;print_newline (); *)
            {
              home_pos = (x,y);
              home_packages = None;
            } 
        ) init.init_homes;
        dx := init.init_dx;
        dy := init.init_dy;
        num := init.init_num;
        x := init.init_x;
        y := init.init_y;
        total_weight := init.init_weight;
        packages := Array.create_matrix (!dx+2) (!dy+2) (None : package list option);
        List.iter (fun (num, x,y) ->
            let r = new_robot num in
            r.robot_pos <- (x,y);
        ) init.init_players

(* This function is called every round, to compute the next command to be 
sent to the server *)
      let play (state : state) = 
(*        Printf.printf "play..."; print_newline ();  *)
        let rec iter () =
(*          Printf.printf "iter"; print_newline (); *)
          match !current_state with
          
          | Looking_for_packages ->
(*              Printf.printf "Looking_for_packages"; print_newline (); *)
              begin
                let best_home = ref None in
                let best_distance = ref max_int in
                List.iter (fun home ->
                    match home.home_packages with
                      Some [] -> ()
                    | _ ->
                        let d = dist (!x, !y) home.home_pos in
                        if d < !best_distance then begin
                            best_distance := d;
                            best_home := Some home;
                          end
                ) !homes;
                match !best_home with
                | Some home ->
                    current_state := MoveToHome home;
                    iter ()
                | None ->
                    try
                      Hashtbl.iter (fun _ p ->
                          match p.pack_final, p.pack_pos with
                          | Some _, Some _ -> 
                              current_state := MoveToPackage p;
                              raise Exit
                          | _  -> ()
                              
                      ) all_packages;
                      Hashtbl.iter (fun _ p ->
                          match p.pack_pos with
                            None -> ()
                          | Some _ ->
                              current_state := MoveToPackage p;
                              raise Exit
                      ) all_packages;
                      Drop []
                      
                    with Exit -> iter ()
              end
          
          | MoveToHome home ->
(*              Printf.printf "MoveToHome %d %d" home.home_x home.home_y; print_newline (); *)
              let hx,hy = home.home_pos in
              
              if (!x,!y) = home.home_pos then begin
                  home.home_packages <- Some (state.state_location_packages);
                  current_state := Looking_for_packages;
                  iter ()
                  end else
              if List.mem home.home_pos 
                [(!x,!y-1); (!x, !y+1); (!x-1,!y); (!x+1,!y)] then
                try
                  Hashtbl.iter (fun _ r ->
                      if home.home_pos = r.robot_pos then
                        raise Exit
                  ) robots;
                  best_move_to hx hy
                with Exit -> Drop []
              else
                best_move_to hx hy
          
          | MoveToPackage p ->
(*              Printf.printf "MoveToPackage %d %b" p.pack_num p.pack_delivered; 
              print_newline (); *)
              match p.pack_pos with
                None -> 
                  current_state := Looking_for_packages;
                  iter ()
              | Some (px,py) ->
                  if px = !x && py = !y then
                    if p.pack_delivered then begin
                        Hashtbl.remove all_packages p.pack_num;
                        current_state := Looking_for_packages;
                        iter ()
                      end else begin
(*                        Printf.printf "At expected position of package"; 
                        print_newline (); *)
                        current_state := Looking_for_packages;
                        iter ()                        
                      end
                  else
                    best_move_to px py
        
        in
        update_state state;
(*        Printf.printf "Score: %d " !score; print_newline (); 
        (match !current_packages with
            { pack_final = Some (px, py) } :: _ -> 
              Printf.printf "State: %d (to %d x %d) packages %d/%d weight"
                (List.length !current_packages)  px py 
                !current_weight !total_weight;
              print_newline (); 
          | _ -> Printf.printf "No packages"; print_newline ());
*)        
        let expected_move =  
          match !current_packages with
            { pack_final = Some (px, py) } :: _ ->
              px, py          
          | _ -> !x, !y
        in
        
        match can_drop () with
        | (x :: _) as list -> 1, Drop list
        |  _ -> match can_pick state expected_move with
            | (x :: _) as list -> 1, Pick list
            | _ -> 
                match can_kill () with
                | r :: _ ->
                    let rx, ry = r.robot_pos in
                    1, best_move_to rx ry
                | _ ->
                    match !current_packages with
                      { pack_final = Some (px, py) } :: _ ->
                        1, best_move_to px py
                    | _ -> -1, iter ()
                        
                
                
(* This function is called if there is some time left before the reply from
the server to  the command *)
      let busy_before_reply = None
        
(* This function is called if there is some time left before the packages
at the current location are received from the server *)
      let busy_before_packages = None

    end)
