
type case =
  Empty
| Wall
| Sea
| Home
| HomeUnchecked
| EmptyUnchecked
| SeaUnchecked
  
type direction = N | E | S | W

type package = {
    pack_num : int;
    mutable pack_weight : int;
    mutable pack_final : (int * int) option;
    mutable pack_pos : (int * int) option;
    mutable pack_delivered : bool;
  }
  
type command =
  Move of direction
| Pick of package list
| Drop of package list
| XY of int * int (* not a command, but a move *)
  
type init = {
    mutable init_map : case array array;
    mutable init_num : int;
    mutable init_weight : int;
    mutable init_money : int;
    mutable init_nplayers : int;
    mutable init_players : (int * int * int) list;
    mutable init_dx : int;
    mutable init_dy : int;
    mutable init_x : int;
    mutable init_y : int;
    mutable init_nhome : int;
    mutable init_nsea : int;
    mutable init_nempty : int;
    mutable init_homes : (int * int) list;
  }

type state = {
    mutable state_map : case array array;
    mutable state_location_packages : package list;
    mutable state_moves : (int * (command list)) list;
  }

type robot = {
    robot_num : int;
    mutable robot_pos : int * int;
    mutable robot_alive : bool;
    mutable robot_packages : package list;
  }
  
let all_packages = Hashtbl.create 113
let new_package num =
  try Hashtbl.find all_packages num with _ ->
      let p = {
          pack_num = num;
          pack_weight = min_int;
          pack_pos = None;
          pack_final = None;
          pack_delivered = false;
        } in
      Hashtbl.add all_packages num p;
      p
 
let robots = Hashtbl.create 13
let new_robot robot_num =
  try Hashtbl.find robots robot_num with _ ->
      let r = {
          robot_num = robot_num;
          robot_pos = (1,1);
          robot_alive = true;
          robot_packages = [];
        } in
      Hashtbl.add robots robot_num r;
      r
      
type final = {
    final_pos : int * int;
    mutable final_after : bool;
    mutable final_used : bool;
    mutable final_packages : package list;
    mutable final_weight : int;
    mutable final_dist : int;
  }
  
type home = {
    home_pos : int * int;
    mutable home_packages : package list option;
  } 
